// Truevision Targa file format specifications: http://www.dca.fee.unicamp.br/~martino/disciplinas/ea978/tgaffs.pdf

#include "../CFile.h"
#include <stdlib.h>
#include "TGA.h"

TGA tga;

void TGA::printError(const char* error, CFile::SFile* f)
{
    printf("TGA %s : %s\n", f->fileName, error);
    file.close(f);
    exit(EXIT_FAILURE);
}

Color* TGA::load(const char* filename, int* width, int* height)
{
    CFile::SFile*   f = file.open(filename, "rb");

    // read header
    SHeader header;
    readHeader(f, &header);

    // skip image ID
    if (header.IDLength != 0)
        file.seek(f, header.IDLength, SEEK_CUR);

    // read palette
    Color*  palette = readPalette(f, &header);

    // read image
    Color* image = readImage(f, &header, palette);
    *width  = header.imageWidth;
    *height = header.imageHeight;

    if (palette != NULL)
        delete[] palette;

    file.close(f);
    return image;
}

void TGA::readHeader(CFile::SFile* f, SHeader* h)
{
    h->IDLength = file.readU8(f);

    h->colorMapType = file.readU8(f);
    if (h->colorMapType > 1)
        printError("Wrong format", f);

    h->imageType = file.readU8(f);

    if (h->imageType != eIndexed &&
        h->imageType != eTrueColor &&
        h->imageType != eGray &&
        h->imageType != eIndexedRLE &&
        h->imageType != eTrueColorRLE &&
        h->imageType != eGrayRLE)
    {
        printError("Wrong format", f);
    }

    h->colorMap_firstEntryIndex = file.readU16(f);
    h->colorMap_length = file.readU16(f);
    h->colorMap_entrySize = file.readU8(f);

    if (h->colorMapType == 1 && h->colorMap_length != 0)
    {
        if (h->colorMap_entrySize != 15 &&
            h->colorMap_entrySize != 16 &&
            h->colorMap_entrySize != 24 &&
            h->colorMap_entrySize != 32)
        {
            printError("Wrong format", f);
        }
    }

    h->xOrigin = file.readU16(f);
    h->yOrigin = file.readU16(f);
    h->imageWidth  = file.readU16(f);
    h->imageHeight = file.readU16(f);
    h->pixelDepth = file.readU8(f);

    if (h->pixelDepth != 8 &&
        h->pixelDepth != 16 &&
        h->pixelDepth != 24 &&
        h->pixelDepth != 32)
    {
        printError("Wrong format", f);
    }

    Uint8 flags = file.readU8(f);

    if ((flags & 0xC0) != 0)
        printError("Wrong format", f);

    h->alphaBits = flags & 0xF;

    if (h->alphaBits != 0 &&
        h->alphaBits != 1 &&
        h->alphaBits != 8)
    {
        printError("Wrong format", f);
    }

    if ((flags & 0x10) == 0)
        h->hDirection = eLeftToRight;
    else
        h->hDirection = eRightToLeft;

    if ((flags & 0x20) == 0)
        h->vDirection = eBottomToTop;
    else
        h->vDirection = eTopToBottom;

    if (h->imageType == eIndexed && h->colorMap_length == 0)
        printError("Wrong format", f);

    if (h->imageType == eIndexedRLE && h->colorMap_length == 0)
        printError("Wrong format", f);
}

Color* TGA::readPalette(CFile::SFile* f, SHeader* h)
{
    Color*  palette = NULL;
    int bytesPerColor = (h->colorMap_entrySize + 7) / 8;

    if (h->imageType == eIndexed || h->imageType == eIndexedRLE)
    {
        palette = new Color[h->colorMap_length];

        for (int i = 0; i < h->colorMap_length; i++)
        {
            if (h->colorMap_entrySize == 15 || h->colorMap_entrySize == 16)
            {
                Uint16 value = file.readU16(f);
                palette[i].b = ( value        & 0x1F) << 3;
                palette[i].g = ((value >>  5) & 0x1F) << 3;
                palette[i].r = ((value >> 10) & 0x1F) << 3;
                palette[i].a = 255;
            }
            else if (h->colorMap_entrySize == 24)
            {
                palette[i].b = file.readU8(f);
                palette[i].g = file.readU8(f);
                palette[i].r = file.readU8(f);
                palette[i].a = 255;
            }
            else if (h->colorMap_entrySize == 32)
            {
                palette[i].b = file.readU8(f);
                palette[i].g = file.readU8(f);
                palette[i].r = file.readU8(f);
                palette[i].a = file.readU8(f);
            }
        }
    }
    else
    {
        file.seek(f, h->colorMap_length * bytesPerColor, SEEK_CUR);
    }

    return palette;
}

Color*  TGA::readImage(CFile::SFile* f, SHeader* h, Color* palette)
{
    int bytesPerPixel = (h->pixelDepth + 7) / 8;
    Color*  image = new Color[h->imageWidth * h->imageHeight];
    Uint8*  imageData = new Uint8[h->imageWidth * h->imageHeight * bytesPerPixel];
    int pixelIndex = 0;

    if (h->imageType == eIndexed ||
        h->imageType == eTrueColor ||
        h->imageType == eGray)
    {
        file.read(f, imageData, h->imageWidth * h->imageHeight * bytesPerPixel);
    }
    else
    {
        // compressed image
    }

    for (int y = 0; y < h->imageHeight; y++)
        for (int x = 0; x < h->imageWidth; x++)
        {
            Color   c;

            // read the next pixel
            Uint32 pixel = 0;

            for (int i = 0; i < bytesPerPixel; i++)
                pixel |= imageData[pixelIndex++] << (i * 8);

            // decode color
            if (h->imageType == eIndexed || h->imageType == eIndexedRLE)
            {
                c = palette[pixel - h->colorMap_firstEntryIndex];
            }
            else if (h->imageType == eGray || h->imageType == eGrayRLE)
            {
                int g = pixel >> (h->pixelDepth - 8);
                c = Color(g, g, g, 255);
            }
            else
            {
                if (bytesPerPixel == 2)
                {
                    c.b = ( pixel        & 0x1F) << 3;
                    c.g = ((pixel >>  5) & 0x1F) << 3;
                    c.r = ((pixel >> 10) & 0x1F) << 3;
                    c.a = 255;
                }
                else if (bytesPerPixel == 3)
                {
                    c.b =  pixel        & 0xFF;
                    c.g = (pixel >>  8) & 0xFF;
                    c.r = (pixel >> 16) & 0xFF;
                    c.a = 255;
                }
                else if (bytesPerPixel == 4)
                {
                    c.b =  pixel        & 0xFF;
                    c.g = (pixel >>  8) & 0xFF;
                    c.r = (pixel >> 16) & 0xFF;
                    c.a = (pixel >> 24) & 0xFF;
                }
            }

            // handle image directions
            int destX, destY;

            if (h->hDirection == eLeftToRight)
                destX = x;
            else
                destX = h->imageWidth - 1 - x;

            if (h->vDirection == eTopToBottom)
                destY = y;
            else
                destY = h->imageHeight - 1 - y;

            image[destY * h->imageWidth + destX] = c;
        }

    return image;
}
